unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls,
  Menus, BaseUnix, TermIO, Linux;

///////////////////////////////////////////////////////////////////////////////////
// remember to add -dUseCThreads to Project -> Project Options -> Custom Options //
///////////////////////////////////////////////////////////////////////////////////

// for use of inotify routines, see:
// https://www.freepascal.org/~michael/articles/dirwatch/dirwatch.pdf

type

  { TForm1 }

  TForm1 = class(TForm)
    Memo1: TMemo;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    Panel1: TPanel;
    PopupMenu1: TPopupMenu;
    Timer1: TTimer;
    procedure FormActivate(Sender: TObject);
    procedure MenuItem1Click(Sender: TObject);
    procedure MenuItem2Click(Sender: TObject);
    procedure MenuItem3Click(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }


const CLlength=20;
var ChangeList:array[0..CLlength-1] of record
                                         mask:integer;
                                         name:string
                                       end;
const clHead:integer=0;
      clTail:integer=0;

 type TCheckThread = class(TThread)
     private
     protected
       procedure Execute; override;
     end;

// separate thread used to check for changes in /dev
procedure TCheckThread.Execute;
var fd, wd, bytesread, filenamelength:integer;
     fds : Tfdset;
     buffer:array [0..4096] of byte;
     event : ^inotify_event;
     filename:string;
     p:pchar;
begin
  fd:=inotify_init();
  try
    wd:=inotify_add_watch(fd, '/dev', IN_CREATE or IN_DELETE {or IN_ISDIR or IN_ONLYDIR});
    fpFD_Zero(fds);
    fpFD_SET(fd, fds);

    while (fpSelect(fd+1, @fds, nil, nil, nil)>=0) do
    begin
      bytesread:=fpRead(fd, buffer, sizeof(buffer));
      event:=@buffer;

      While ((pchar(event)-@buffer)<bytesread) do
      begin
        filenamelength:=event^.len;

        if (filenamelength>0) then
        begin
          p:=@event^.name+filenamelength-1;
          While (p^=#0) do
          begin
            dec(p);
            dec(filenamelength)
          end
        end;

        setlength(filename, filenamelength);
        if (filenamelength>0) then move(event^.name,filename[1],filenamelength);
                                   //     source    destination      size

        if pos('tty',filename)=1 then
//      if event^.wd=wd then
        begin
          ChangeList[clHead].mask:=event^.MASK;                        // valid port names are saved in a queue
          ChangeList[clHead].name:=filename;                           // ... ready to be picked up by the timer
          clHead:=(clHead+1) mod CLlength;                             // ... that runs every 100ms.
        end;

//      writeln('Change 0x', IntToHex(event^.mask, 8),
//              ' detected for "/dev/', filename, '"');

        ptruint(event):=ptruint(event)+sizeof(inotify_event)+event^.len-1
      end
    end;
  finally
    fpClose(fd)
  end
end;






type
   TSerialStruct = packed record
          typ: cint;
          line: cint;
          port: cuint;
          irq:  cint;
          flags: cint;
          xmit_fifo_size: cint;
          custom_divisor: cint;
          baud_base: cint;
          close_delay: cushort;
          io_type: cchar;
          reserved_char:  pcchar;
          hub6: cint;
          closing_wait: cushort; // time to wait before closing
          closing_wait2: cushort; // no longer used...
          iomem_base: pcchar;
          iomem_reg_shift: cushort;
          port_high: clong; // cookie passed into ioremap
   end;

// method based on information obtained from the following two sites:
// https://www.lazarusforum.de/viewtopic.php?p=72837
// https://stackoverflow.com/questions/2530096
function CheckDevice(DeviceName:string):boolean;               // checks to see if the device named is a live
var DriverName:string;                                         // ... serial port. this is done by checking
            FD:longint;                                        // ... entries in /sys/class/tty including the
            SS:TSerialStruct;                                  // ... presence of a link to a device driver.
            ST:stat;
begin
  Result:=false;

  st.st_mode:=0;
  if (DeviceName<>'.') and (DeviceName<>'..') then
  if FileExists('/sys/class/tty/'+DeviceName+'/device/driver') then
  if fpLstat('/sys/class/tty/'+DeviceName+'/device', ST)=0 then
  if fpS_ISLNK(ST.st_mode) then
  begin
    DriverName:=ExtractFileName(fpReadLink('/sys/class/tty/'+DeviceName+'/device/driver'));
//  writeln(DeviceName,'  :  ',DriverName);

    if DriverName<>'serial8250' then Result:=true
                                else begin
//                                     writeln('    checking serial8250 device');
                                       FD:=fpOpen('/dev/'+DeviceName, O_RDWR or O_NONBLOCK or O_NOCTTY);
                                       if FD>0 then
                                       try
                                         if fpIOCtl(FD, TIOCGSERIAL, @SS)<>-1 then
                                         if SS.typ<>0 then Result:=true;
                                         fpclose(FD)
                                       except end
                                     end
  end
end;





procedure PopulatePortList(var Memo:TMemo);
var S,S1,S2:string;
      I,J,K:integer;
       flag:boolean;
         SR:TSearchRec;
begin
  Memo.Clear;

  if FindFirst('/sys/class/tty/*', faDirectory , SR) = 0 then          // initially we scan /sys/class/tty for potential serial ports
  repeat                                                               //                   ~~~~~~~~~~~~~~
    if CheckDevice(SR.Name) then Memo.Lines.Add(SR.Name)               // check that each port found is live and add to memo of it is
  until FindNext(SR) <> 0;
  FindClose(SR);

  if Memo.Lines.Count=0 then Memo.Lines.Add('no serial ports found');

  flag:=true;
  while (Memo.Lines.Count>1) and flag do                               // excessively complicated sort routine, tries to ensure that the
  begin                                                                // 'fixed' serial ports appear first, and that port numbers are
    flag:=false;                                                       // ordered correctly: 0,1,2...,8,9,10,11, etc.
    for I:=0 to Memo.Lines.Count-2 do
    begin
      S1:=Memo.Lines[I];
      J:=1+length(S1);

      S2:=Memo.Lines[I+1];
      K:=1+length(S2);

      if (J-K)<0 then begin                                                    // pack S1 with zeros to left of numeric part
                        while (J>1) and (S1[J-1] in ['0'..'9']) do dec(J);
                        while length(S1)<length(S2) do insert('0', S1, J)
                      end else
      if (K-J)<0 then begin                                                    // pack S2 with zeros to left of numeric part
                        while (K>1) and (S2[K-1] in ['0'..'9']) do dec(K);
                        while length(S2)<length(S1) do insert('0', S2, K)
                      end;

      J:=pos('ttyS',S1);                                               // =1 if is a 'fixed' serial port
      K:=pos('ttyS',S2);                                               // =1 if is a 'fixed' serial port

      if ((J<>1) and (K=1)) or                                         // bubble ttyS* ports to top of the list
         ((J=K) and (S1>S2)) then                                      // within respective groups sort alphabetically
      begin
        Memo.Lines.Exchange(I, I+1);
        flag:=true                                                     // flag set if at least one swap during this pass
      end
    end
  end;

  S:=Memo.Text;
  if RightStr(S,1)=#10 then S:=copy(S, 1, length(S)-1);                // remove final linefeed from memo
  Memo.Text:=S;                                                        // (stops the caret going beyond the last line)

  Memo.SelLength := 0;
  Memo.SelStart := Length(Memo.Text)
end;





// 100ms timer that picks up port names from the queue, checks that they are live, and if so adds them into the list held in Memo1.
procedure TForm1.Timer1Timer(Sender: TObject);
var I:integer;
    S:string;
begin
  while clHead<>clTail do                                              // it is extremely unlikely this will loop through more than once
  begin
    Memo1.Lines.BeginUpdate;
    I:=Memo1.Lines.IndexOf('no serial ports found');
    if I>=0 then Memo1.Lines.Delete(I);
    I:=Memo1.Lines.IndexOf(ChangeList[clTail].name);
    if I>=0 then Memo1.Lines.Delete(I);                                // delete the port that changed from the list if it is already there

    if (ChangeList[clTail].mask and IN_CREATE)>0 then                  // if event was port insert then
        if CheckDevice(ChangeList[clTail].name) then                   // ... if is a 'live' serial port
            Memo1.Lines.Add(ChangeList[clTail].name);                  // ... add port to end of the list
    if Memo1.Lines.Count=0 then Memo1.Lines.Add('no serial ports found');

    S:=Memo1.Text;
    if RightStr(S,1)=#10 then S:=copy(S, 1, length(S)-1);              // remove final linefeed from memo
    Memo1.Text:=S;                                                     // (stops the caret going beyond the last line)

    Memo1.Lines.EndUpdate;
    Memo1.SelLength:=0;
    Memo1.SelStart:=Length(Memo1.Text);                                // make sure caret is sitting at the end of the list

    clTail:=(clTail+1) mod CLlength
  end
end;





// one-time startup code, including a full scan of /sys/class/tty and check of every current entry to create an initial list of live serial ports.
procedure TForm1.FormActivate(Sender: TObject);
const initial:boolean=true;
begin
  if initial then
  begin
    initial:=false;
    PopulatePortList(Memo1);

    Form1.Constraints.MinWidth:=Form1.Width;
    Form1.Constraints.MaxWidth:=Form1.Width;
    Form1.Constraints.MinHeight:=Form1.Height;
    Form1.Constraints.MaxHeight:=Form1.Height;

    Timer1.Enabled:=true;
    TCheckThread.Create(false)                                         // hereafter we just scan /dev/tty* for changes
  end                                                                  //                        ~~~~~~~~~
end;                                                                   // all changes are added to the end of the list, no more sorting






procedure TForm1.MenuItem1Click(Sender: TObject);
begin
  Memo1.Clear
end;

procedure TForm1.MenuItem2Click(Sender: TObject);
begin
  PopulatePortList(Memo1)
end;

procedure TForm1.MenuItem3Click(Sender: TObject);
begin
  Form1.Close
end;





end.
